window.uiHtml = new (function(){

  // let self_close_pattern = new RegExp(`^\\<(img|br|hr)`);
  let self_close_pattern = null;
	
	let div = document.createElement('div');

  let allowed_tag_chars = '[a-zA-Z0-9\-_]';
  let allowed_comp_chars = '[a-zA-Z0-9\-_:]';
  let allowed_var_chars = '[a-zA-Z0-9\-_"]';
	let atc = allowed_tag_chars;
	let acc = allowed_comp_chars;
	let avc = allowed_var_chars;

  let close_pattern = new RegExp(`^\\<\\/:?${acc}+>`);
  let loop_pattern = new RegExp(`^\\<(${atc}+)\\[(${avc}*)\\]`);
  let tag_pattern = new RegExp(`^\\<(${atc}+)(\\[(${avc}*)\\])?`);
  let component_pattern = new RegExp(`^\\<:(${acc}+)(\\[(${avc}*)\\])?`);
	let var_pattern_escaped = new RegExp(`^\\{!(${avc}+)(!)\\}`);
  let var_pattern = new RegExp(`^\\{(${avc}+)\\}`);
  let cond_pattern = new RegExp(`^\\<\\?(!?)(${avc}+)(=?)(${avc}+)?`);

  function match(string, regex, matches, i){
    string = string.substr(i);
    let result = string.match(regex);
    matches.splice(0, matches.length, ...(result||[]));
    return result!=null;
  }
	function html_escape(str){
		div.textContent = str;
		return div.innerHTML;
	}

  this.render = function(component, vars=[], is_html=false){
    try{
      let html = this.renderHTML(component, vars, is_html);
      let template = document.createElement('template');
      template.innerHTML = html;
      return template.content;
    }catch(e){
      console.error(e);
    }
  };
  this.renderHTML = function(component, vars=[], is_html=false){
    let file;
    if(is_html == false){
      let ui = document.getElementById('ui:' + component.replace(/\//g, ':'));
      if(ui == null) throw `Component '${component}' does not exist.`;
      file = ui.textContent;
    }else{
      file = component;
    }
    let len = file.length;

    // Aux Variables
    let i = 0;
    let start = 0;
    let string = "";
    let matches = [];

    // Parse templating language
    for(i = 0; i < len; i++){
      if(file[i] == '<'){
        if(match(file, component_pattern, matches, i)){ // Subcomponent tag: <:component>
          let cmp = fetchHtml(file, i, true);
          string += file.substring(start, i);
          start = i = cmp.end;
          i--; // Substract 1 because of iteration incrementation
          let cmp_html = cmp.string;

          let cmp_name = matches[1];
          if(matches[2]){
            (vars[matches[3]]||[]).forEach((_loop, _index, _arr) => {
	            let loop_var = _loop;
	            if(typeof loop_var !== "object") loop_var = {_value: loop_var};
              let _vars = Object.assign({}, vars, loop_var);
              _vars = parseComponentVars(cmp_html, _vars);
              _vars['_index'] = _index;
              _vars['_index_'+matches[3]] = _vars['_index'];
              _vars['_last'] = _arr.length;
              _vars['_last_'+matches[3]] = _vars['_last'];
              let result = this.renderHTML(cmp_name, _vars);
              string += result;
            });
          }else{
            let _vars = parseComponentVars(cmp_html, vars);
            let result = this.renderHTML(cmp_name, _vars);
            string += result;
          }

        }else if(match(file, loop_pattern, matches, i)){ // Looping tag: <div[list] ...>
          let loop = fetchHtml(file, i);
          string += file.substring(start, i);
          start = i = loop.end;
          i--; // Substract 1 because of iteration incrementation
          let loop_html = loop.string;

          loop_html = loop_html.replace(loop_pattern, '<$1');

          let counter = 0;
          let loop_vars = vars[matches[2]]||{};
          for(let j in loop_vars){
            let loop_var = loop_vars[j];
            if(typeof loop_var !== "object") loop_var = {_value: loop_var};
            let _vars = Object.assign({}, vars, loop_var);
            _vars['_index'] = counter++;
            _vars['_index_'+matches[2]] = _vars['_index'];
            _vars['_last'] = loop_vars.length;
            _vars['_last_'+matches[2]] = _vars['_last'];
            let result = this.renderHTML(loop_html, _vars, true);
            string += result;
          }

        }else if(match(file, cond_pattern, matches, i)){ // Conditional: <?var <div> ?>
          let cond = fetchConditional(file, i);
          string += file.substring(start, i);
          let cond_html = cond.string.substring(matches[0].length, cond.string.length - 2);
          let cond_var = vars[matches[2]];
          let cond_var2 = false;
          if(matches[4]){
            cond_var2 = matches[4]||false;
            if(cond_var2 && cond_var2[0]=='"' && cond_var2[cond_var2.length-1]=='"'){
              cond_var2 = matches[4].replace(/^"(.+)"/, '$1');
            }else if(cond_var2){
              cond_var2 = vars[matches[4]]
            }
          }

          let add = "";
          if(matches[3] == "="){
            if((cond_var == cond_var2) ^ (matches[1] == "!")){
              add = cond_html;
            }
          }else if((matches[1] == "!") ^ (!!cond_var)){
            add = cond_html;
          }

          file = file.substring(0, i) + add + file.substring(i+cond.string.length);
          start = i--;
          len = file.length;
        }else if(file.substr(i, 4) == '<!--'){ // Comment: <!-- comment -->
          let comment = fetchComment(file, i);
          string += file.substring(start, i);
          start = i = comment.end;

          let comment_str = comment.string.substring(4, comment.string.length-3).trim();
          if(comment_str == 'UI::END') break;
          i--; // Substract 1 because of iteration incrementation
        }
      }else if(file[i] == '{'){
        if(match(file, var_pattern_escaped, matches, i) || match(file, var_pattern, matches, i)){ // Variable: {var}
          let _var = matches[1];
          let result = vars[_var]||"";
					if(matches[2]) result = html_escape(result);
          string += file.substring(start, i) + result;
          start = i = i + matches[0].length;
          i--; // Substract 1 because of iteration incrementation
        }
      }
    }
    string += file.substring(start, i);
    return string;
  };
  this.renderParts = function(string){
    let matches = [];
    let parts = [];
    let start = 0;
    let len = string.length;
    let i = 0;
    for(i = 0; i < len; i++){
      if(string[i] == '<'){
        if(match(string, component_pattern, matches, i)){ // Subcomponent tag: <:component>
          let cmp = fetchHtml(string, i);
          parts.push({type: 'text', string: string.substring(start, i)});
          start = i = cmp.end;
          i--;
          parts.push({type: 'ui', string: this.renderHTML(cmp.string, [], true)});
        }else if(string.substr(i, 4) == '<!--'){ // Comment: <!-- comment -->
          let comment = fetchComment(string, i);
          parts.push({type: 'text', string: string.substring(start, i)});
          start = i = comment.end;
          i--;

          let comment_str = comment.string.substring(4, comment.string.length-3).trim();
          if(comment_str == 'UI::END') break;
        }
      }
    }
    parts.push({type: 'text', string: string.substring(start, i)});

    return parts;
  };

  let parseComponentVars = function (string, vars){
    let depth = 0;
    let matches = [];
    let len = string.length;
    let return_vars = {};

    let no_vars = true;
    for(let i = 0; i < len; i++){
      if(match(string, tag_pattern, matches, i)){
        no_vars = false;
        let _var = matches[1];
        let inner = fetchHtml(string, i, true);

        if(matches[2]){
          let loop_var = matches[3];
          if(loop_var === ""){
            let result = parseComponentVars(inner.string, vars);
            if(!return_vars[_var]) return_vars[_var] = [];
            return_vars[_var].push(result);
          }else{
            let counter = 0;
            (vars[loop_var]||[]).forEach((loop, _index, _arr) => {
	            let loop_var = loop;
	            if(typeof loop_var !== "object") loop_var = {_value: loop_var};
              let _vars = Object.assign({}, vars, loop_var);
              _vars['_index'] = _index;
              _vars['_index_'+matches[3]] = _vars['_index'];
              _vars['_last'] = _arr.length;
              _vars['_last_'+matches[3]] = _vars['_last'];
              let result = parseComponentVars(inner.string, _vars);
              if(!return_vars[_var]) return_vars[_var] = [];
              return_vars[_var].push(result);
            });
          }
        }else{
          return_vars[_var] = this.renderHTML(inner.string, vars, true);
        }
        i = inner.end - 1;
      }else if(string[i] == '>' && string[i-1] == '/'){
        i++;
        if(--depth == 0) break;
      }
    }

    if(no_vars) return this.renderHTML(string, vars, true);
    return return_vars;
  }.bind(this);
  function fetchHtml(string, start, discard_first){
    let depth = 0;
    let matches = [];
    let len = string.length;
    let i = start;

    let first_start = 0;
    let first_end = 0;
    let self_close = false;

    for(i = start; i < len; i++){
      if(string[i] == '<'){
        if(self_close_pattern && match(string, self_close_pattern, matches, i)){
          self_close = true;
        }

        if(match(string, close_pattern, matches, i)){
          let i_before = i;
          i += matches[0].length - 1;
          depth--;
          if(depth == 0){
            i++;
            first_end = i_before;
            break;
          }
        }else if(string.substr(i, 4) == '<!--'){
          let comment = fetchComment(string, i);
          i = comment.end;
        }else{
          depth++;
        }
      }else if(string[i] == '>' && (string[i-1] == '/' || string[i-1] == '?' || self_close)){
        self_close = false;
        depth--;

        if(depth == 1){
          first_end = i;
        }else if(depth == 0){
          i++;
          break;
        }
      }else if(string[i] == '>' && depth == 1){
        first_start = i+1;
      }
    }

    if(discard_first) string = string.substring(first_start, first_end);
    else string = string.substring(start, i);

    return {
      'string': string,
      'end': i
    };
  }
  function fetchComment(string, start){
    let depth = 0;
    let len = string.length;
    let i = start;

    for(i = start; i < len; i++){
      if(string[i] == '>' && string.substr(i-2, 3) == '-->'){
        i++;
        break;
      }
    }
    return {
      'string': string.substring(start, i),
      'end': i
    };
  }
  function fetchConditional(string, start){
    let depth = 0;
    let len = string.length;
    let i = start;
    let matches = [];

    for(i = start; i < len; i++){
      if(string[i] == '<' && match(string, cond_pattern, matches, i)){
        depth++;
      }if(string[i] == '>' && string[i-1] == '?' && 1==depth--){
        i++;
        break;
      }
    }
    return {
      'string': string.substring(start, i),
      'end': i
    };
  }

})();
